#ifndef BOOTSTRAP_SERVER_H
#define BOOTSTRAP_SERVER_H

#include<vector>


class CFileTransfer;

/**
* Bootstrap Server class.
* The server listens on specified port as listed in the configuration file.
* The server maintains a file that lists a bulk of IP:Port that can be used
* for bootstraping for kademlia nodes.
*/

class CBootStrapServer{


public:
	/**
	* Constructor of the class. 
	* It takes configuration file name as its argument.
	* An example of such configuration file is:
        * \verbinclude server_config.txt
	*/
	CBootStrapServer(const std::string& sConfigFile);
	
	/**
	* Destructor of the class.
	*/
	virtual ~CBootStrapServer();

	/**
	* This method is used to start the server.
	*/
	int Start();

	/**
	* This method is used to stop the server.
	*/
	int Stop()const;

private:

	/**
	* In this method, the server waits in a poll for any incomming connection from clients.
	*/
	int Run();

	/**
	* Once server receives a connection request and accepts it, this method is used to handle that connection.
	* It sends a SMIME signed file to the client listing the possible ip:port of nodes for bootstrapping.
	*/
	int HandleConnection(int iFd);

	/**
	* Handles the notification that is sent by the FileTransfer class after sending the file.
	* Also used to stop the main loop of the server when server wants to shutdown.
	*/
	int HandleNotification(int iFd);
	int CreateDaemonSocket();

	/**
	* Loads file that lists the ip:port of possible bootstraping nodes.
	*/
	int LoadNodes();
	int InitSocketPair();

	/**
	* Parses the configuration file. 
	* Returns -1 if an misconfiguration found in the configuration file.
	*/
	int ParseConfig();
	void StrTrim(std::string& str);
	void DestroySocket();
	int GetFileSize(const std::string& sFileName)const;
	std::string GetString(const std::vector<std::string> & vConf, const std::string& sLine) const;
	/**
	* Signs the node file in SMIME format with the private key and also attaches servers certificate in the signed SMIME file.
	*/
        int SignNodeFile();
	void SearchAndDeleteFileObj(const CFileTransfer* obj);
	void StopPollLoop()const;


private:
	std::string m_sConfigFile;
	std::string m_sNodeFile;
	std::string m_sSignedNodeFile;
	int m_iDaemonSockFd;
	int m_iSockPairFd[2];
	uint32_t m_uServerIP;
	uint16_t m_uServerPort;
	std::string m_sServerPrivKey;
	std::string m_sServerCert;
	std::string m_sLogFile;
	std::vector<uint32_t>m_vNodes;

	typedef std::vector<CFileTransfer*> FileTransfer;
	FileTransfer m_filetrans;
	bool m_bRun;


};

#endif//BOOTSTRAP_SERVER_H

