#ifndef FILE_TRANSFER_H
#define FILE_TRANSFER_H

#include "../aMule-2.2.4/src/MD4Hash.h"

#define MAXSIZE         1024

/**
* CFileTransfer class.
* It is used to make any file transfer between two nodes over TCP.
* This class is used for different file tranfer such as:
* Kad file tranfer, gossip file transfer, missing gossip file transfer, pollution attack robustness file trnafer etc.
*/
class CFileTransfer{

public:
	/**
	* Constructor. It takes the socket as its argument. Because, before creating an object, initiator node
	* needs to call static function ConnectPeer function. If that is successful, it returns the socket fd, 
	* which is used in this argument. SharedDir is used to download certificate files which have been searched
	* from p2p KAD network. Download directory is used to download Gossip Messages.
	*/
	CFileTransfer(int iSockFd, const std::string& sSharedDir, const std::string& sDownloadDir, int iSearchID=-1);

	/**
	* Destructor of the class.
	*/
	virtual ~CFileTransfer();

	/**
	* A static function to connect peer using iRemoteIP and iRemotePort. 
	* If the connectionis successful, then the socket fd is returned, else -1 is returned.
	*/
	static int ConnectPeer(uint32_t iRemoteIP, uint16_t iRemotePort);

	int GetSockFd(){return m_iSockFd;}
	std::string GetFileName(){return m_sFileName.c_str();}
	int GetFileSize(){return m_iFileSize;}

	int GetSearchID(){return m_iSearchID;}

	void SetSearchID(int iSearchID){ m_iSearchID = iSearchID;}

	std::string GetSharedDir(){return m_sSharedDir;}
	std::string GetDownloadDir(){return m_sDownloadDir;}
	void Cleanup();
	void CloseSocket();

	/**
	* File Hash is stored so that later the corresponding object can be retrived.
	*/
	void SetFileHash(const CMD4Hash& fHash){ m_fileHash = fHash;}
	CMD4Hash GetFileHash(){return m_fileHash;}
	void SetSocketPair(int iFd){m_iSocketPairFd = iFd;}

	/**
	* GOssip ID
	*/
	void SetGossipIDCheck(bool b){ m_bCheckGossipId = b;}

	void SetGossipID(const std::string& id){ m_sGossipId = id;}
	std::string GetGossipID(){return m_sGossipId;}
	void SetGossipList(const std::vector<std::string>& v){ m_vGossipList = v;}
	std::vector<std::string> GetGossipList(){return m_vGossipList;}


	/**
	* A socket pair is used to communicate between sender/receiver threads and the main thread.
	*/
	int GetSocketPair(){return m_iSocketPairFd;}

	/**
	* Sends an EventMessage to the network thread after file tranfer is done.
	*/
	void NotifyComplete(bool bSuccess);

	/**
	* File tranfer is used both gossiping (push based) and certificate download using kademlia (pull based).
	* So, to make the sender and receiver dynamic, this function is used instead of static sender or receiver.
	* HandShake funtion sends EventMessage to the peer. After handshaking, sender and receiver becomes fixed.
	*/
	int HandShake(bool bPassive, bool bPush=false, const std::string& sFileName=std::string(), int iFileSize=0);
	int HandShake(bool bPassive, bool bPush, int iSockFd);
	uint8_t GetFileType(){return m_8FileType;}
	void SetFilePath(const std::string& sPath){m_sFilePath = sPath;}
	bool IsCompleted(){return m_bCompleted;}
	const std::string GetExtraData(){return m_sExtraData;}

	/**
	* Starts the file tranfer thread.
	*/
	int StartFileTransfer();
	void SetRemoteAddress(uint32_t iRemoteIP, uint16_t iRemotePort) {m_iRemoteIP=iRemoteIP; m_iRemotePort=iRemotePort;}

private:
	int SendPacket(int iSockFD, const char* szPayload, int iSize); /*code is unused at this time*/
	
	/**
	* Parses the Handshake EventMessage to figure out who is sender, who is receiver of a particular file.
	*/
	int ParseHandShakePacket(const char* szBuffer, int iReceived, bool& bPull);

	/**
	* Waits for handshake packet in the poll.
	*/
	int WaitForHandShakePacket(bool& bPull, bool bWaitForAck=false);

	/**
	* Waits for handshake packet Acknowledgement in the poll.
	*/
	int WaitForHandShakePacketAck();

	/**
	* Sends the handshake packet.
	*/
	int SendHandShakePacket(bool bPush);
	int StartFileTransferThread();
	static void* FileTransferThread(void* arg);

	/**
	* Receives a file.
	*/
	int ReceiveFile();

	/**
	* Sends a file.
	*/
	void SendFile();
	void NotifySenderReport(bool bSuccess);

private:
	int m_iSockFd;
        std::string m_sFileName;
        std::string m_sFilePath;
        int m_iFileSize;
	pthread_t m_pTID;
	int m_iSearchID;
	std::string m_sSharedDir;
	std::string m_sDownloadDir;
	CMD4Hash m_fileHash;
	int m_iSocketPairFd;
	uint8_t	m_8FileType;
	std::string m_sExtraData;
	bool m_bCompleted;
	bool m_bPassive;
	bool m_bPush;
	std::string m_sGossipId;
	bool m_bCheckGossipId;
	std::vector<std::string> m_vGossipList;
 	uint32_t m_iRemoteIP; 
	uint16_t m_iRemotePort;


};


#endif //FILE_TRANSFER_H
