#ifndef GOSSIP_H 
#define GOSSIP_H 



#include <boost/uuid/uuid.hpp>
#include <boost/uuid/uuid_generators.hpp>

#include "kademlia_interface.h"
#include "EventMessage.h"



#define GOSSIP_PROTO    0xC1
#define MAXCHAR        1500
#define MAX_FORWARD     3

#define MAX_GOSSIP_SEND	5

/**
* This structure is used to save gossip message according to time
* so that if any neighbor doesn't receive a particular gossip message,
* that can be sent to him.
*/

struct _ThreadPoolEntry{
	const std::string m_sFileName; 
	uint32_t m_iDestinationHost; 
	uint16_t m_iDestinationPort;
	boost::uuids::uuid m_Tag;
	_ThreadPoolEntry(const std::string& sFile, uint32_t iIP, uint16_t iPort, boost::uuids::uuid id):
		m_sFileName(sFile),
		m_iDestinationHost(iIP),
		m_iDestinationPort(iPort),
		m_Tag(id)
		{}
};
typedef struct _ThreadPoolEntry ThreadPoolEntry;

struct _TransactionLog{

	time_t t; 
	boost::uuids::uuid Tag;
	std::string sFileName;
	_TransactionLog(time_t time, boost::uuids::uuid id, const std::string& s):
		t(time),
		Tag(id),
		sFileName(s){
	}
};
typedef _TransactionLog TransactionLog;


/**
* Only used in UDP gossiping.
*/
struct gossipPacket{
        uint8_t m_iProtocol;
        boost::uuids::uuid m_Tag;
	/*We put command packet as  a stream..it save time for packet converstion */
        uint8_t m_cPack[MAXCHAR];
        //CommandPacket m_cPack;
};
typedef struct gossipPacket GossipPacket;

typedef std::map<boost::uuids::uuid, std::vector< std::string > > MapFwHist;
typedef std::map<boost::uuids::uuid, std::vector< Kademlia::CContact* > > MapNeighbor;

class CBootStrapClient;


/**
* CGossip class. It extends functionality of CKadInterface class
* since the gossiping works on top of kademlia protocol.
*/
class CGossip:public CKadInterface{

public:
	/**
	* CGossip Destructor.
	*/
	virtual ~CGossip();


	/**
	* This method is used to get an instance of CGossip class.
	* It is used to make this architecture as a singleton pattern.
	*/
	static CGossip* getGossipInstance(std::string);

	/**
	* Overrides the functionality of NotifyApp of CKadInterface to send 
	* EventMessage to the Policy Component.
	*/
	void NotifyApp(const std::string& sFileName, uint8_t filetype, uint8_t relatedModule);
	void NotifyZeroContact();
	void NotifyAllContacts(const std::string& sList);


	//void HandleSenderReport();
	void HandleSenderReport(const std::string& sFileName, bool b);


	/**
	* It is used to get the log of already received Gossip Message Transaction ID
	* to attached in the KAD PING message.
	*/
	int GetTransactionLog(std::string& transList)const;

	/**
	* From the KAD PING message, this method is used to find if the node missed any gossip message.
	*/
	int FindMissingGossip(const std::string& sTransList, std::string& sMissingList)const;

	/**
	* If any missing Gossip Message is found, the transaction is written in the KAD PONG
	* message to retrieve it.
	*/
	int SendMissingGossip(const std::string& sMissingList, uint32_t ip, uint16_t port);



private:
	static bool bInstance;
	static CGossip* instance;
	/**
	* CGossip Constructor; Takes configuration file name as its argument.
	* It is private due to make the class as a singleton pattern.
	*/
	CGossip(std::string sFileName);

private:
	/**
	* Process any received UDP Gossip Message.
	*/
        void ProcessGossipPacket(const GossipPacket& gPack, uint32 ip, uint16 port, size_t length);

	/**
	* Send a gossip message using UDP protocol.
	*/
        int SendGossipPacket(const GossipPacket& packet, int pktLen, uint32_t destinationHost, uint16_t destinationPort);

	/**
	* This method is invoked after receiving any Kademlia message.
	*/
	void OnPacketReceived(uint32 ip, uint16 port, byte* buffer, size_t length);

	/**
	* This method is invoked after receiving EventMessage from Policy Component.
	*/
        int OnCommandReceived(int iFd);

	/**
	* It sends a UDP Gossip Message.
	*/
        int SendCommand(const GossipPacket& gPack, int len, uint32_t srcIP, uint16_t srcPort);

	/**
	* It sends a TCP Gossip Message.
	* Tag is used as a unique ID for this gossip message.
	*/
	int SendTcpCommand(const std::string& sFileToSend, const boost::uuids::uuid& Tag, uint32_t srcIP, uint16_t srcPort, int iCricLevel);

	/**
	* Sends a gossip message in the network.
	*/
        int SendCommand(const std::string& sCommand, bool bTcp = false, uint8_t module = MODULE_GOSSIP);

	/**
	* This actually sends data to the Policy Component.
	*/
	int NotifyCommand(const std::string&);

	/**
	* Get the transaction id as UUID from sFile. 
	* sFile is supposed to be in SMIME format and id should be found in "FROM:" field.
	*/
	int GetTransactionID(const std::string& sFile, boost::uuids::uuid& id, int& iCricLevel);

	/**
	* Sends a gossip message over TCP transport protocol.
	*/
	int SendTcpGossipPacket(const std::string& sFileName, uint32_t destinationHost, uint16_t destinationPort, const boost::uuids::uuid& id);

	/**
	* Process any received TCP Gossip Message.
	*/
	void ProcessTcpGossipPacket(const std::string& sFileName, uint32 ip, uint16 port, uint8_t module);
	std::string CopyFileToTmpDir(const std::string & sFile);

	/**
	* Invokes Bootstrap Client for getting bootstrap IP and port.
	*/
	int GetKadBootstrapInfo(int iServerIP, int iPort);

	/**
	* After receiving or sending any TCP Gossip message, the message id, time and the file name is saved 
	* so that id list can be sent to the neighbor in PING message.
	*/
	int DoTransactionLog(const  boost::uuids::uuid&Tag, const std::string& sFileName);
	std::string BackupFileToDownloadDir(const std::string & sFile, const  boost::uuids::uuid& Tag);

	/**
	* If a neighbor sends a gossip message ID in PONG message, the message is sent to the neighbor.
	*/
	int SendMissingGossip(const boost::uuids::uuid& Tag, uint32_t ip, uint16_t port);

	//std::string GetGMTime(void);

 
	int ServiceThreadPool();
	void SubstituteFailedNodes(const std::string& sFileName);
	int HandleCriticalLevel(int iCricLevel);



private:
        /*Gossip*/
        MapFwHist mFwHist;
        MapFwHist mRecvHist;

        typedef std::deque<CFileTransfer*> FileTransfer;
        FileTransfer m_gfiletrans;
        CBootStrapClient* m_pBootStrapClient;

	//gossip transaction logging
	std::deque<TransactionLog> m_qTransactionLog;
	std::deque<ThreadPoolEntry> m_qPoolEntry;
	unsigned int m_iThreadCount;

        MapNeighbor mNeighbor;


};

#endif //GOSSIP_H


