#ifndef KADEMLIA_INTERFACE_H 
#define KADEMLIA_INTERFACE_H 

//namespace Kademlia_interface{
#include "kademlia/kademlia/Kademlia.h"
#include "kademlia/kademlia/Prefs.h"
#include "Logger.h"
#include "kademlia/utils/KadUDPKey.h"
#include "multicast.h"



class CKnownFileList;
class CSharedFileList;
class CSearchList;
class CSearchFile;
class CPartFile;
class CFileTransfer;
class CMulticast;

typedef std::pair<uint32_t, uint16_t> AddrPortPair;

/**
* An class for interfacing with kademlia protocol.
* It is also main class for communication component.
*/
class CKadInterface{

public:

	/**
	* Constructor of the class. 
	* Takes configuration file name of the communication component as its argument.
	*/
	CKadInterface(std::string sFilename);

	/**
	* Destructor of the class.
	*/
	virtual ~CKadInterface();

	/**
	* Main function to run the communication component.
	* Returns -1 on failure in running the component.
	* It runs Kademlia, multicast and network modules.
	*/
	int Run();

	/**
	* Asks kademlia to startup.
	*/
	void StartKad();

	/**
	* Asks kademlia to stop.
	*/
	void StopKad();

	/**
	* Asks kademlia to bootstrap on some specific ip and port in kad network.
	*/
	void BootstrapKad();

	/**
	* Retrieves contact list of neighbor from kademlia protocol.
	* This contact list is useful for gossiping.
	*/
	void GetAllContacts() ;
	void FindNodeByID(const uint8_t* nodeID);
	uint32_t GetPublicIP(bool ignorelocal) ;
	int GetSockFd(){return m_iSockFd;}
	int GetDaemonSockFd(){return m_iDaemonSockFd;}
	int GetClientTcpSockFd(){return m_iClientTcpSockFd;}
	int GetPeerSockFd(){return m_iPeerSockFd;}
	void SetPeerSockFd(int iFd){m_iPeerSockFd=iFd;}

	/**
	* Cleans up any allocated resource such as sockets, threads, etc.
	*/
	void CleanUp();

	/**
	* Search source of a file in the KAD network.
	*/
	void SearchSrcKad(CSearchFile* toadd);

	/**
	* This function is called after a source of a file is found.
	*/
	void KademliaSearchFile(uint32_t searchID, const Kademlia::CUInt128* pcontactID, const Kademlia::CUInt128* pbuddyID, uint8_t type, uint32_t ip, uint16_t tcp, uint16_t udp, uint32_t buddyip, uint16_t buddyport, uint8_t byCryptOptions);
	CPartFile* GetFileByID(const CMD4Hash& filehash) const;

	/**
	* Retrieves the name of the shared directory. 
	* Kademlia publishes files who are located in this shared directory.
	* After searching a file in the kad network and downloading it, 
	* the file is also stored in this shared directory for further publishing.
	*/
	std::string GetSharedDir(){return m_sSharedDir;}

	/**
	* Retrieves the download directory. 
	* Any gossip file is stored in this directory.
	*
	*/
	std::string GetDownloadDir(){return m_sDownloadDir;}
	const int* GetSocketPair(){return m_iSockPairFd;}

	/**
	* Asks kademlia to publish the sFile in the KAD network.
	*/
	void PublishFile(const std::string& sFile);

	/**
	* Asks kademlia to search any keyword in the kad network.
	*/
	void SearchKeyKad(const std::string& sSearchStr);

	/**
	* Stores the source ip and port for a specific kad search ID.
	*/
	void AddSource(uint32_t searchID, uint32_t ip, uint16_t port);

	/**
	* After file transfer notifies the network thread, this function checks the file transfer status result
	* and invokes corresponding action.
	*/
	int CheckFileTransferStatus(int iFd);
	std::string MakeAbsPath(const std::string& sPath);

	/**
	* Sets the path of the download directory as specified in the configuration file.
	*/
	void SetDownloadDir(const std::string& sPath);

	/**
	* Returns the multicasting instance.
	*/
	CMulticast* GetMulticastInstance(){return m_pMulticast;}	

	/**
	* Checks whether the gossip module is enabled or not.
	* In the configurationfile, it is possible to enable or disable gossip module and multicast module.
	*/
	bool IsGossipEnable(){return m_bGossipEnable;}

	/**
	* Checks whether the multicast module is enabled or not.
	* In the configurationfile, it is possible to enable or disable gossip module and multicast module.
	*/
	bool IsMulticastEnable(){return m_bMulticastEnable;}

	/**
	* Asks kademlia to bootstrap the specified ip and port.
	*/
	void BootstrapKad(uint32_t ip, uint16_t port);





private:
	/**
	* Starts the networking thread. 
	* It listens for any :
	* Kademlia message,
	* Gossip file transfer message,
	* socket pair for communicating with the main thread,
	* policy component message
	*/
	void StartNetworkThread();
	void SetLogger();
	static void *ThreadCallback(void *arg);

	/**
	* Creates a UDP socket for kademlia.
	*/
	int CreateSocket();

	/**
	* Creates a TCP socket for communicating with policy component.
	*/
	int CreateDaemonSocket();

	/**
	* Creates a TCP socket for kademlia.
	*/
	int CreateClientTcpSocket();

	void CheckEvent(int iFd);
	int CheckDaemonEvent(int iFd);
	void DestroySocket();

	/**
	* Parse the configuration file.
	* Returns -1 on error in the configuration file.
	*/
	int ParseConfig();
	std::string GetString(const std::vector<std::string>& vConf, const std::string& sLine);
	void StrTrim(std::string&)const;

	/**
	* Executes on timeout in the network thread. 
	* The interval is 1 sec.
	* Calls Kademlia::Process()
	*/
	void OnTimer();
	void SetSharedDir(const std::string& sPath);

	/**
	* Invoked after a Kademlia packet is arrived.
	*/
	virtual void OnPacketReceived(uint32_t ip, uint16_t port, byte* buffer, size_t length);

	/**
	* Invoked after a command from policy component is received.
	* Override by the CGossip class.
	* Communication is done using EventMessage class.
	*/
	virtual int OnCommandReceived(int iFd){std::cout<<"i am in kad\n"; return 0;}

	/**
	* It sends EventMessage to the Policy component. 
	* sFileName, filetype and relatedModule are encoded in the EventMessage.
	*/
	virtual void NotifyApp(const std::string& sFileName, uint8_t filetype, uint8_t relatedModule );

	/**
	* After receiving a Gossip Message on TCP, this method is invoked.
	*/
	virtual void ProcessTcpGossipPacket(const std::string& sFileName, uint32 ip, uint16 port, uint8_t module);

	CPartFile* GetFileByKadFileSearchID(uint32 id) const;
	CFileTransfer* CheckFileTransByID(const int iSearchID) const;
	CFileTransfer* CheckFileTransByHash(const CMD4Hash& fHash) const;
	int InitSocketPair();

	/**
	* If a file tranfer is failed for downloading a file from the KAD network,
	* retry is done by invoking this function.
	*/
	void RetrySource(int iSearchID);

	/**
	* Creates Shared Dir, Download Dir and .tmp Dir. 
	*/
	void MakeDirs();
	void ShutdownNetworkThread();


private:
	int m_iSockFd;
	int m_iDaemonSockFd;
	int m_iPeerSockFd;
	int m_iClientTcpSockFd;

	std::string m_sConfigFileName;
	std::string m_sClientName;
	bool m_bBootstrapEnable;
	bool m_bKadLogEnable;
	uint32_t m_uClientIP;
	uint16_t m_uClientPort;
	uint32_t m_uBootIP;
	uint16_t m_uBootPort;
	pthread_t m_Thread;
	uint16_t m_uDaemonTcpPort;
	uint16_t m_uClientTcpPort;

        CKnownFileList* knownfiles; 

        typedef std::deque<CPartFile*> FileQueue;
        FileQueue m_filelist;

	std::string m_sSharedDir;
	std::string m_sDownloadDir;
	int m_iSockPairFd[2];

	std::map<int, std::vector<AddrPortPair> > m_SourceMap;

	//multicast
	bool m_bMulticastEnable;
	bool m_bGossipEnable;
	std::string m_sMulticastInterface;
	uint32_t m_uMulticastIP;
	uint16_t m_uMulticastPort;

	CMulticast *m_pMulticast;
	bool	m_bInit;

public:
        typedef std::deque<CFileTransfer*> FileTransfer;
        FileTransfer m_filetrans;

        CSharedFileList* sharedfiles; 
        CSearchList* searchlist ;
};




//}//end of Kademlia_interface

#endif //KADEMLIA_INTERFACE_H 

