#ifndef GOSSIP_COMM_H 
#define GOSSIP_COMM_H 

#include<vector>
#include<map>
#include<deque>

#include <boost/uuid/uuid.hpp>
#include <boost/uuid/uuid_generators.hpp>

#include "digiSig.h"

#define MAX_SIZE	1500

typedef uint8_t           byte;
typedef std::map<std::string, uint8_t> RcvdFileMap;

class CDigiSig;

/**
* This is the main class for policy module that communicates with the "Communication" module.
* It also uses SMIME for sign or verify the gossip commands.
*/


/** Placement of Policy component in TCP/IP stack
*  \image html c2_system_arch.jpg "C2 System Architecture"
*  \image latex c2_system_arch.eps "C2 System Architecture" width=10cm
*/


class CGossipComm{

public:
	/**
	* Constructor of the CGossipComm class; it takes configuration file name as its argument
	* An example of such configuration file is:
	* \verbinclude policy_config.txt
	*/
        CGossipComm(std::string sFileName);
        virtual ~CGossipComm();
	/**
	* Main method for running this policy module.
	* Returns -1 if it is unable to run for configuration error in the configuration file.
	*/
	int Run();

	/**
	* Runs the network thread; The thread waits for any communication from the Communication module.
	*/
	void StartNetworkThread();
	/**
	* Sends the command written the file sCommand to the Communication module.
	* Before sending the command, it signs the command file using SMIME.
	* For signing, private key and the certificate are required which are searched to the location 
	* specified in the configuration file.
	* Returns -1 if fails.
	*/
        int SendCommand(const std::string& sCommand);
	/**
	* It calls int SendCommand(const std::string& sCommand) by putting m_sCommandFileName as its argument.
	* Returns -1 if fails.
	*/
        int SendCommand();
	/**
	* Asks Communication module for finding sFile using Kademlia protocol.
	*/
	int FindFile(const std::string& sFile);
	/**
	* Loads the subscription file in the run time. The name of the subscription file is specified in the configuration file.
	* The module will execute commands which are from the nodes listed in the subscription file. If "All" is specified in the
	* subscription file, then all commands will be executed regardless of the senders.
	*/
	bool LoadSubscription();

	/**
	* A socket pair is created to communicate between the main thread and the network thread. If the main thread exists, it writes
	* shutdown on this socket pair so that network thread shutdowns and thus a graceful shutdown takes place.
	*/
        const int* GetSocketPair(){return m_iSockPairFd;}
        const int* GetLogSocketPair(){return m_iLoggerSockPairFd;}
        int InitLoggerSocketPair();
	int StartLogger();
	int StopLogger();
	void Logging(const std::string& sLog);



private:
	/**
	* Main function that is run by the network thread.
	*/
        static void *ThreadCallback(void *arg);
	
        int GetSockFd(){return m_iSockFd;}
        void OnCommandReceived(byte* buffer, size_t length);
        void CheckEvent(int iFd);
	/**
	* sFileName file is sent to the Communication module for gossiping. 
	* module refers to the corresponding module such as MULTICAST or GOSSIP etc. 
	*/
	int SendToGossip(const std::string& sFileName, uint8_t module);

	/**
	* A TCP socket is created to talk to the Communication module.
	*/
	int CreateSocket();

	/**
	* Bootstrap node's IP and port can be known from the central bootstrap server.
	* To find such a bootstrap IP and port, the function asks the Communication module
	* to communicate with the bootstrap server and receive ip and port.
	*/
	void FindBootStrapNode();
	
	/**
	* Parses the configuration file. 
	* Returns -1 if there is an error in the configuration file.
	*/
        int ParseConfig();
	std::string GetString(std::vector<std::string> & vConf, const std::string& sLine);
	std::string GetTwoSidedString(std::vector<std::string>& vConf, const std::string& sLine, std::string& sLeft);
	void StrTrim(std::string& str)const;

	/**
	* Verifies SMIME signature in the sFileName file.
	*/
	bool VerifySignature(const std::string& sFileName);

	/**
	* Asks Communication module to publish the file sCert in the p2p network.
	*/
	int PublishCert(const std::string& sCert);
	/**
	* Copies 'sFile' to the Key directory and returns the path of the copied file.
	* The name of the key directory is specified in the configuration file.
	* Key directory generally means a certificate cache directory.
	*/
	std::string CopyFileToKeyDir(const std::string& sFile);
	/**
	* Copies 'sFile' to the command directory. Generally all command files are stored in this directory.
	*/
	std::string CopyFileToCommandDir(const std::string & sFile);
	std::string MakeAbsPath(const std::string& sPath);
	/**
	* Asks the Communication module to publish CA certificates. It is usefull to publish any CA certificates.
	*/
	void PublishCACerts();
	std::string GetCAName(const std::string& sCert);
	bool FileExistsInKeyDir(const std::string& sFileName);
	/**
	* Verifies a certificate in the certificate chain.
	* Returns false if varification fails.
	*/
	bool VerifyCA();
	/**
	* From a certificate, if the policy component tries to find the issuer certificate. If the issuer certificate is not found
	* in the certificate cache directore, then it calls FindFile(const std::string&) function.
	*/
	bool CollectCerts(std::string& sFileName);

	/**
	* After receiving all relevent certificates, it verifies the whole certificate chain. 
	* Returns true if verification becomes successfull.
	*/
	bool ReverseVerification(bool bVerifyCA=true);
	std::string GetPath(const std::string& sFileName);
	std::string GetCertificateName(const std::string& sCert);

	/**
	* Checks if the Issuer name is the trusted higher CA or not.
	*/
	bool IsSignedByHigherCA(const std::string& sIssuer);
	void DeleteFileFromCache(const std::string& sPath);



	/*Subscription handlers*/
	/**
	* Checks whether the policy component has subscribed to the sFrom. 
	*/
	bool IsSubscribed(const std::string& sFrom);
	/**
	* If verification is successfull, then policy component gossips the command by sending the signed file to the Communication component and 
	* Executes the command if the command has arrived for the first time.
	*/
	void GossipAndExecute(const std::string& sFileName, const std::string& sFileToExecute);

	bool IsFirstTimeReceived(const std::string& sFile, std::string& sSub);
        int InitSocketPair();
	void ShutdownNetworkThread();

	/**
	* Once received authenticated IP and port from bootstraping server, policy component asks the Communication module for doing Kademlia Bootstrapping.
	*/
	void BootStrapKadReq(int ip, int port);
	uint8_t CheckFileType(const std::string& sFile);
	int GetKadBootIPandPort(const std::string& sFile);
	void GetAddrPortFromString(const std::string& sLine, int& ip, int& port)const;
	bool CheckSelfReceived(const std::string& sFrom)const;
	/**
	* Find the sensitivity level from the sender's name.
	*/
	int ExtractRunLevel(const std::string& sFrom)const;
	/**
	* Check whether the sender's name in his certificate and his anounced name in sFrom are same.
	*/
	bool CheckSenderIntigrity(const std::string& sFile, const std::string& sCert);

	/**
	* Load the policy profile which gives a map between the sensitivity level and the corresponding parameter.
	*/
	int LoadPolicyProfile();
	uint8_t CheckModuleType(const std::string& sFile);

	/**
	* If a certificate is received, it is also sent back to the communication component. 
	* this is necessary only for pollution attack robustness technique.
	*/
        int ReportFeedback(const std::string& sFileName, bool bSuccess);

	/**
	* Creates command_store repository and certificate_cache_dir repository 
	* if don't exist.
	*/
	void MakeDirs();

	/**
	* Communication with Log Server..
	*/
	int SendToLogServer(const std::string& sLog);
	int SendLog( const std::string& sLog);
	int CreateLogSocket();
	void DestroyLogSocket();
	int ReceiveAck();
	void StartLoggerThread();
	static void* LoggerThreadCallback(void *arg);
	std::string GetGMTime(void);
	int GetRandomIndex();





private:
	/**
	* Configuration file name
	*/
	std::string	m_sConfigFile;
        pthread_t m_Thread;
        int m_iSockFd;
	/**
	* The IP address on which Communication module is listening.
	*/
	uint32_t m_uGossipIP;
	/**
	* The port on which Communication module is listening.
	*/
        uint16_t m_uGossipPort;
	uint32_t m_uClientIP;
        uint16_t m_uClientPort;
	//std::string m_sPrivateKey;
	std::string m_sTestCertificate;//only for testing

	/**
	* Certificate cache directory.
	*/
	std::string m_sKeyDir;
	//std::string m_sCertificate;
	bool	m_bSignEnable;

	/*We may need to remove the boost id, since it may no longer useable*/
	/**
	* The node's Identity
	*/
        boost::uuids::uuid m_ClientName;
        std::string m_sClientName;

	/**
	* A Map for certificate name and its issuer name relationship.
	*/
	std::map<std::string, std::string> m_packet;
	/**
	* Vector for CA chain hierarchy.
	*/
	std::vector<std::string> m_vCAs;

	/*Subscription variables*/
	std::vector<std::string> m_vSubscribed;
	std::string m_sSubscriptionFileName;
	bool m_bSubscribedAll;

	/**
	* Name of the trusted CA.
	*/
	std::string m_sTrustedCAName;
	/**
	* Path of the trusted CA file.
	*/
	std::string m_sTrustedCAPath;
	std::string m_sCommandRepository;
	std::string m_sLogFileName;
	std::string m_sCommandFileName;

	std::map<std::string, std::string> m_cert;
	/**
	* An object for Openssl interface for SMIME authentication.
	*/
	CDigiSig m_digisig;
	std::map<std::string, int> m_ca_in_file;

	std::set<std::string> m_issuer_set;
	/**
	* Supports both TCP and UDP gossip. If UDP gossip, then does not attach the senders certificate in the SMIME message.
	*/
	int m_iTransportProtocol;

	std::map<std::string, int> m_MessageIDMap;
        int m_iSockPairFd[2];
	bool m_bInit;

	//bootstrap info from bootstrap server
	int m_iBootServerIP;
	int m_iBootServerPort;

	std::map<std::string, uint8_t> m_recv_file_types;
	bool	m_bBootstrapEnable;

	//multiple user ids
	std::map<std::string, std::string> m_UserId;
	std::string m_sDefaultUserId;
	std::string m_sClientCertDir;
	std::string m_sPolicyProfile;

	/**
	* Maps the sensitivity level and corresponding parameter in the profile.
	*/
	std::map<int,std::string> m_ProfileMap;
	std::map<std::string, uint8_t> m_recv_module_types;

	//only for validating certificates that are being tried to publish by peer
	bool m_bValidateCert;
	std::deque<std::string> m_vCert;
	
	//for log communication
        int m_iLogSockFd;
	int m_iLogServerIP;
	int m_iLogServerPort;
        pthread_t m_LoggerThread;
        int m_iLoggerSockPairFd[2];
	bool m_bLoggerInit;

	typedef std::pair<int, int> AddrPortPair;
	std::vector<AddrPortPair> m_vBootAddrPort;
	int m_iBootIndex;

	bool m_bFloodingEnable;
	bool m_bNonCooperationEnable;

	std::string m_sCricLevel;

};



#endif //GOSSIP_COMM_H


